/*
Copyright (c) 2010, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <grub/cpu/io.h>
#include "pci.h"

#define PCI_ADDR_PORT 0xCF8
#define PCI_DATA_PORT 0xCFC

static U32 pci_legacy_addr(U8 bus, U8 dev, U8 fn, U8 reg)
{
    return (1 << 31) | (bus << 16) | (dev << 11) | (fn << 8) | reg;
}

U8 pci_read_u8(U8 bus, U8 dev, U8 fn, U32 reg)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    return grub_inb(PCI_DATA_PORT + (reg & 3));
}

U16 pci_read_u16(U8 bus, U8 dev, U8 fn, U32 reg)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    return grub_inw(PCI_DATA_PORT + (reg & 3));
}

U32 pci_read_u32(U8 bus, U8 dev, U8 fn, U32 reg)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    return grub_inl(PCI_DATA_PORT + (reg & 3));
}

void pci_write_u8(U8 bus, U8 dev, U8 fn, U32 reg, U8 value)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    grub_outb(value, PCI_DATA_PORT + (reg & 3));
}

void pci_write_u16(U8 bus, U8 dev, U8 fn, U32 reg, U16 value)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    grub_outw(value, PCI_DATA_PORT + (reg & 3));
}

void pci_write_u32(U8 bus, U8 dev, U8 fn, U32 reg, U32 value)
{
    grub_outl(pci_legacy_addr(bus, dev, fn, reg) & ~3, PCI_ADDR_PORT);
    grub_outl(value, PCI_DATA_PORT + (reg & 3));
}

static U32 pci_mmcfg_addr(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U16 reg)
{
    return mmcfg_base | (bus << 20) | (dev << 15) | (fn << 12) | reg;
}

U8 pci_mmcfg_read_u8(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg)
{
    U8 *regAddr;

    regAddr = (U8 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    return *(volatile U8 *)regAddr;
}

U16 pci_mmcfg_read_u16(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg)
{
    U16 *regAddr;

    regAddr = (U16 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    return *(volatile U16 *)regAddr;
}

U32 pci_mmcfg_read_u32(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg)
{
    U32 *regAddr;

    regAddr = (U32 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    return *(volatile U32 *)regAddr;
}

void pci_mmcfg_write_u8(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg, U8 value)
{
    U8 *regAddr;

    regAddr = (U8 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    *regAddr = value;
}

void pci_mmcfg_write_u16(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg, U16 value)
{
    U16 *regAddr;

    regAddr = (U16 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    *regAddr = value;
}

void pci_mmcfg_write_u32(U32 mmcfg_base, U8 bus, U8 dev, U8 fn, U32 reg, U32 value)
{
    U32 *regAddr;

    regAddr = (U32 *) pci_mmcfg_addr(mmcfg_base, bus, dev, fn, reg);

    *regAddr = value;
}
